from unittest.mock import patch
from chaosterraform import control


@patch("chaosterraform.control.Terraform", autospec=True)
def test_configure_control_with_parameters(mocked_driver):
    control.configure_control(silent=False, retain=True, chdir="../testfolder/one", configuration={})

    mocked_driver.assert_called_once_with(silent=False, retain=True, chdir="../testfolder/one", args={})
    mock_instance = mocked_driver.return_value
    mock_instance.terraform_init.assert_called()


@patch("chaosterraform.control.Terraform", autospec=True)
def test_configure_control_with_terraform_variables(mocked_driver):
    configuration = {
        "tf__variable_str": "sg-00001111",
        "tf__one_more_bool_var": True,
        "tf__numeric": 1.2,
    }

    control.configure_control(configuration=configuration)

    expected_variables = {
        "variable_str": "sg-00001111",
        "one_more_bool_var": True,
        "numeric": 1.2,
    }

    mocked_driver.assert_called_once_with(silent=True, retain=False, chdir=None, args=expected_variables)


@patch("chaosterraform.control.Terraform", autospec=True)
def test_configure_control_configuration_override(mocked_driver):
    configuration = {
        "tf_conf__silent": False,
        "tf_conf__retain": True,
        "tf_conf__chdir": "my/other/folder/path",
    }

    control.configure_control(silent=True, retain=False, chdir="some/folder", configuration=configuration)

    mocked_driver.assert_called_once_with(silent=False, retain=True, chdir="my/other/folder/path", args={})


@patch("chaosterraform.control.Terraform", autospec=True)
def test_before_experiment_control_with_outputs(mocked_driver):
    outputs = {
        "dns_name": {"value": "example.com"},
        "desired_count": {"value": 1},
        "cpu_units": {"value": 1024},
    }
    configuration = {}
    mock_instance = mocked_driver.return_value
    mock_instance.output.return_value = outputs
    control.before_experiment_control(context=None, configuration=configuration)

    mock_instance.apply.assert_called_once()
    mock_instance.output.assert_called_once()

    assert configuration.get("tf_out__dns_name") == "example.com"
    assert configuration.get("tf_out__desired_count") == 1
    assert configuration.get("tf_out__cpu_units") == 1024


@patch("chaosterraform.control.Terraform", autospec=True)
def test_after_method(mocked_driver):
    configuration = {}
    mock_instance = mocked_driver.return_value
    mock_instance.retain = False

    control.after_experiment_control(context=None, state=None, configuration=configuration)

    mock_instance.destroy.assert_called_once()


@patch("chaosterraform.control.Terraform", autospec=True)
def test_after_method_retain_resources(mocked_driver):
    configuration = {}
    mock_instance = mocked_driver.return_value
    mock_instance.retain = True

    control.after_experiment_control(context=None, state=None, configuration=configuration)

    mock_instance.destroy.assert_not_called()



"""
Terraform control module

This module allows Chaos Toolkit users to create infrastructure resources using Terraform scripts
for the experiment execution.
"""
from chaoslib.types import Configuration, Experiment, Journal, Secrets, Settings
from logzero import logger

from .driver import Terraform

VAR_NAME_PREFIX = "tf__"
CONFIG_PREFIX = "tf_conf__"
EXPORT_VAR_PREFIX = "tf_out__"


def configure_control(
    silent: bool = True,
    retain: bool = False,
    chdir: str = None,
    configuration: Configuration = None,
    secrets: Secrets = None,
    settings: Settings = None,
    experiment: Experiment = None,
):
    """
    Configure terraform control for the experiment execution

    Parameters
    ----------
    silent: bool
        suppress Terraform logs in ChaosToolkit experiment logs
    retain: bool
        retain created resources at the end of the experiment
    chdir: str
        change the Terraform working directory

    Raises
    ------
    InterruptExecution
        If terraform init fails we interrupt the experiment execution immediately
    """
    # pylint: disable=unused-argument
    tf_vars = {}
    if configuration:
        for key, value in configuration.items():
            if key.startswith(VAR_NAME_PREFIX):
                tf_key = key.replace(VAR_NAME_PREFIX, "")
                tf_vars[tf_key] = value

    for key, _ in tf_vars.items():
        configuration.pop(f"{VAR_NAME_PREFIX}{key}")

    params = {
        "retain": bool(configuration.get(f"{CONFIG_PREFIX}retain", retain)),
        "silent": bool(configuration.get(f"{CONFIG_PREFIX}silent", silent)),
        "chdir": configuration.get(f"{CONFIG_PREFIX}chdir", chdir),
    }
    logger.info(
        "Terraform: retain stack after experiment completion: %s",
        str(params.get("retain")),
    )

    driver = Terraform(**params, args=tf_vars)
    driver.terraform_init()


def before_experiment_control(
    context: Experiment,
    configuration: Configuration = None,
    secrets: Secrets = None,
    **kwargs,
):
    """
    before-control of the experiment's execution

    Apply the Terraform stack before the experiment execution. As the experiment did not start
    yet, if the resources creation fails the execution is interrupted immediately.

    Raises
    ------
    InterruptExecution
        interrupts the experiment execution if resources creation fails
    """
    # pylint: disable=unused-argument
    driver = Terraform()
    logger.info("Terraform: creating required resources for experiment")
    driver.apply()
    for key, value in driver.output().items():
        logger.info("Terraform: reading configuration value for [%s]", key)
        configuration[f"{EXPORT_VAR_PREFIX}{key}"] = value.get("value")


def after_experiment_control(
    context: Experiment,
    state: Journal,
    configuration: Configuration = None,
    secrets: Secrets = None,
    **kwargs,
):
    """
    after-control of the experiment's execution

    Destroy resources after the experiment execution unless retain specifically requested by
    the experiment using the `retain` parameter
    """
    # pylint: disable=unused-argument
    driver = Terraform()
    if not driver.retain:
        logger.info("Terraform: removing experiment resources")
        driver.destroy()
    else:
        logger.info("Terraform: stack resources will be retained after experiment completion.")



""" A Chaos Toolkit driver to run Terraform commands """
import functools
import json
import os
import subprocess
from copy import deepcopy
from itertools import chain
from typing import Dict, List

from chaoslib.exceptions import InterruptExecution
from logzero import logger


def _run(*cmd: List[List[str]], capture_output: bool = False, text: bool = False):
    _cmd = list(chain(*cmd))
    return subprocess.run(_cmd, shell=False, capture_output=capture_output, text=text, check=False)


def singleton(cls):
    """A singleton wrapper for any class"""

    @functools.wraps(cls)
    def wrapper_singleton(*args, **kwargs):
        if not wrapper_singleton.instance_:
            wrapper_singleton.instance_ = cls(*args, **kwargs)
        return wrapper_singleton.instance_

    wrapper_singleton.instance_ = None
    return wrapper_singleton


@singleton
class Terraform:
    """
    The Terraform driver to run CLI commands from the Python program

    Attributes
    ----------
    retain: bool
        A flag to indicate if created resources should be retained at the end of the experiment.
        By default all resources are destroyed
    silent: bool
        Suppress Terraform output in ChaosToolkit logs
    chdir: str
        A path to the terraform working directory. If the directory does not exists the experiment
        is interrupted with an InterruptExecution exception
    args: Dict
        Terraform variables overrides

    Raises
    ------
    InterruptExecution
        If the chdir path does not exists we interrupt the experiment execution immediately
    """

    def __init__(
        self,
        retain: bool = False,
        silent: bool = False,
        chdir: str = None,
        args: Dict = None,
    ):
        super().__init__()
        self.retain = retain
        self.silent = silent
        self.chdir = chdir
        self.args = args or {}

    @property
    def _terraform(self):
        if self.chdir:
            if not os.path.exists(self.chdir):
                raise InterruptExecution(f"Terraform: chdir [{self.chdir}] does not exists")
            if not os.path.isdir(self.chdir):
                raise InterruptExecution(f"Terraform: chdir [{self.chdir}] is not a directory")
            return ["terraform", f"-chdir={self.chdir}"]
        return ["terraform"]

    def terraform_init(self):
        """
        Initialize Terraform modules

        Raises
        ------
        InterruptExecution
            Interrupts the experiment execution if Terraform modules could not be initialized
        """
        if not os.path.exists(".terraform"):
            result = _run(self._terraform, ["init"], capture_output=self.silent)
            if result.returncode != 0:
                if self.silent:
                    logger.error(result.stderr.decode("utf-8"))
                raise InterruptExecution("Failed to initialize terraform")

    def apply(self, **kwargs):
        """
        Apply the Terraform stack

        Parameters
        ----------
        kwargs: Dict
            Keyword arguments to temporarily override Terraform variables for the apply

        Raises
        ------
        InterruptExecution
            Interrupts the experiment execution if the Terraform stack failed to create
        """

        args = deepcopy(self.args)
        args.update(kwargs)

        var_overrides = []
        for key, value in args.items():
            string_value = value
            if isinstance(value, bool):
                string_value = str(value).lower()
            var_overrides.extend(["-var", f"{key}={string_value}"])

        result = _run(
            self._terraform,
            ["apply", "-auto-approve"],
            var_overrides,
            capture_output=self.silent,
        )
        if result.returncode != 0:
            if self.silent:
                logger.error(result.stderr.decode("utf-8"))
            raise InterruptExecution("Failed to apply terraform stack terraform")

    def output(self):
        """
        Reads Terraform stack outputs into a Python dict

        Returns
        -------
        dict
            A dictionary of output parameters from the Terraform stack
        """
        result = _run(self._terraform, ["output", "-json"], capture_output=True, text=True)
        outputs = json.loads(result.stdout)
        return outputs

    def destroy(self):
        """
        Destroy the resources created by the Terraform stack
        """
        _run(
            self._terraform,
            ["destroy", "-auto-approve"],
            capture_output=self.silent,
        )



from unittest.mock import patch
from chaosterraform import driver


class MockShellResponse:
    def __init__(self, returncode=0, text=""):
        self.returncode = returncode
        self.stdout = text


@patch("chaosterraform.driver.os.path")
def test_chdir(mocked_path):
    mocked_path.exists.return_value = True

    driver.Terraform.instance_ = None
    tf_driver = driver.Terraform(chdir="../testfolder/one")

    assert tf_driver._terraform == ["terraform", "-chdir=../testfolder/one"]


@patch("subprocess.run")
def test_init_default_args(mocked_run):
    mocked_run.return_value = MockShellResponse()
    driver.Terraform.instance_ = None
    tf_driver = driver.Terraform(silent=True)
    tf_driver.terraform_init()

    mocked_run.assert_called_once_with(
        ["terraform", "init"],
        shell=False,
        capture_output=True,
        text=False,
        check=False,
    )


@patch("subprocess.run")
def test_apply_default_args(mocked_run):
    mocked_run.return_value = MockShellResponse()
    driver.Terraform.instance_ = None
    tf_driver = driver.Terraform()
    tf_driver.apply()

    mocked_run.assert_called_once_with(
        ["terraform", "apply", "-auto-approve"],
        shell=False,
        capture_output=False,
        text=False,
        check=False,
    )


@patch("subprocess.run")
def test_apply_verbose(mocked_run):
    mocked_run.return_value = MockShellResponse()
    driver.Terraform.instance_ = None
    tf_driver = driver.Terraform(silent=True)
    tf_driver.apply()

    mocked_run.assert_called_once_with(
        ["terraform", "apply", "-auto-approve"],
        shell=False,
        capture_output=True,
        text=False,
        check=False,
    )


@patch("subprocess.run")
def test_destroy_default_args(mocked_run):
    driver.Terraform.instance_ = None
    tf_driver = driver.Terraform()
    tf_driver.destroy()

    mocked_run.assert_called_once_with(
        ["terraform", "destroy", "-auto-approve"],
        shell=False,
        capture_output=False,
        text=False,
        check=False,
    )


@patch("subprocess.run")
def test_output_strings(mocked_run):
    return_text = '{"expected":"value","one": "1","boolean":true,"number":999}'
    mocked_run.return_value = MockShellResponse(text=return_text)
    driver.Terraform.instance_ = None
    tf_driver = driver.Terraform()
    result = tf_driver.output()

    mocked_run.assert_called_once_with(
        ["terraform", "output", "-json"],
        shell=False,
        capture_output=True,
        text=True,
        check=False,
    )
    assert result.get("expected") == "value"
    assert result.get("one") == "1"
    assert result.get("boolean") is True
    assert result.get("number") == 999



"""Top-level package for chaosterraform"""
from typing import List

from chaoslib.discovery.discover import discover_actions, discover_activities, initialize_discovery_result
from chaoslib.types import DiscoveredActivities, Discovery

name = "chaosterraform"
__author__ = """Manuel Castellin"""
__email__ = "manuel@castellinconsulting.com"
__version__ = "0.0.7"
__all__ = [
    "discover",
    "__version__",
]


def discover(discover_system: bool = True) -> Discovery:
    """Discover chaostoolkit activities"""
    # pylint: disable=unused-argument
    discovery = initialize_discovery_result("chaosterraform", __version__, "chaosterraform")
    discovery["activities"].extend(load_exported_activities())
    return discovery


def load_exported_activities() -> List[DiscoveredActivities]:
    """
    Extract metadata from actions and probes exposed by this extension.
    """
    activities = []
    activities.extend(discover_actions("chaosterraform.actions"))
    activities.extend(discover_activities("chaosterraform.control", "control"))

    return activities
